package frc.robot.subsystems;

import java.util.function.DoubleSupplier;

import com.ctre.phoenix6.hardware.CANcoder;
import com.revrobotics.spark.SparkMax;
import com.revrobotics.spark.SparkLowLevel.MotorType;

import edu.wpi.first.math.controller.ArmFeedforward;
import edu.wpi.first.math.controller.PIDController;
import edu.wpi.first.wpilibj2.command.Command;
import edu.wpi.first.wpilibj2.command.SubsystemBase;
import frc.robot.constants.ArmConstants;

public class Arm extends SubsystemBase {
    private SparkMax armMotor;

    private CANcoder canCoder;

    private PIDController positionController;
    private PIDController velocityController;

    private ArmFeedforward feedForward;

    public Arm() {
        armMotor = new SparkMax(
            ArmConstants.kArmMotorID, 
            MotorType.kBrushless
        );

        canCoder = new CANcoder(ArmConstants.kCANcoderID);
    }

    //manual command that keeps ouput speed consistent no matter the direction
    public Command runArm(DoubleSupplier speed) {
        return run(() -> {
            double realSpeedTarget = speed.getAsDouble() * ArmConstants.kArmMaxVelocity;

            double voltsOut = velocityController.calculate(
                rotationsToRadians(canCoder.getVelocity().getValueAsDouble()), 
                realSpeedTarget
            ) + feedForward.calculate(
                rotationsToRadians(canCoder.getPosition().getValueAsDouble()), 
                canCoder.getVelocity().getValueAsDouble()
            );

            armMotor.setVoltage(voltsOut);
        });
    }

    public Command goToSetpoint(double setpoint, double timeout) {
        return run(() -> {
            double voltsOut = positionController.calculate(
                canCoder.getPosition().getValueAsDouble(), 
                setpoint
            ) + feedForward.calculate(
                canCoder.getPosition().getValueAsDouble(), 
                canCoder.getVelocity().getValueAsDouble()
            );

            armMotor.setVoltage(voltsOut);
        }).until(positionController::atSetpoint).withTimeout(timeout);
    }

    protected double rotationsToRadians(double rotations) {
        return rotations * 2 * Math.PI;
    }
}
